"""Data object for player friends."""
import datetime

from pydantic import BaseModel
from pydantic import Field
from pydantic.types import UUID4

from .utils import to_camel


class Friend(BaseModel):
    """Friend object.

    Args:
        id (bool): Id of friend.
        uuid_sender (UUID4): UUID of player sending friend request.
        uuid_receiver (UUID4): UUID of player receiving friend request.
        started (datetime.datetime): Time players started being friends.
    """

    id: str = Field(alias="_id")
    uuid_sender: UUID4
    uuid_receiver: UUID4
    started: datetime.datetime

    class Config:
        """Config."""

        alias_generator = to_camel



"""Profile objects."""
import datetime
import math
import uuid
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from pydantic import BaseModel


class Quests(BaseModel):
    """Armor.

    Args:
        status (str): Status of quest.
        activated_at (datetime.datetime): Time activated.
        activated_at_sb (datetime.datetime): Time activated.
        completed_at (datetime.datetime): Time Completed.
        completed_at_sb (datetime.datetime): Time Completed.
    """

    status: str
    activated_at: datetime.datetime
    activated_at_sb: datetime.datetime
    completed_at: datetime.datetime
    completed_at_sb: datetime.datetime


class Objective(BaseModel):
    """Armor.

    Args:
        status (str): Status of objective.
        progress (int): Progress through objective.
        completed_at (Optional[datetime.datetime]): Time compelted at. Defaults to None.
    """

    status: str
    progress: int
    completed_at: Optional[datetime.datetime] = None


class InvArmor(BaseModel):
    """Armor.

    Args:
        type (int): Type of armor.
        data (str): Data of armor.
    """

    type: int
    data: str


class Members(BaseModel):
    """Member.

    Args:
        last_save (datetime.datetime):Time last saved.
        inv_armor (InvArmor): Armor.
        first_join (datetime.datetime): Time first joined.
        first_join_hub (Optional[datetime.datetime]): first joined hub.
        stats (Dict[str, int]): Member stats.
        objectives (Dict[str, Objective]): Objectives.
        tutorial (List[str]): Tutorial.
        quests (Dict[str, Quests]): Quests done.
        coin_purse (Optional[int]): Amount of coins in purse.
        last_death (datetime.datetime): Time last died.
        crafted_generators (List[str]): Crafted generators.
        visited_zones (Optional[List[str]]): Visited zones.
        fairy_souls_collected (int): Souls collected.
        fairy_souls (Optional[int]): Fairy souls.
        death_count (Optional[int]): death count.
        slayer_bosses (Dict[str, Dict[str, Any]]): Slayer bosses.
        pets (List[Any]): Pets.
    """

    last_save: Optional[datetime.datetime]
    first_join: Optional[datetime.datetime]
    coin_purse: float = 0
    fairy_souls_collected: int = 0
    fairy_souls: int = 0
    fairy_exchanges: int = 0
    pets: List[Any] = []
    collection: Dict[str, Any] = {}
    collections_unlocked: int = 0

    inv_armor: Optional[InvArmor]
    first_join_hub: Optional[int]
    stats: Optional[Dict[str, float]]
    tutorial: List[str] = []
    last_death: Optional[int]
    crafted_generators: List[str] = []
    visited_zones: List[str] = []
    death_count: int = 0

    @property
    def fairy_bonus(self) -> Dict[str, int]:
        """Bonus from fairy.

        Returns:
            Dict[str, int]: Fairy bonus.
        """
        bonus = {"speed": 0, "strength": 0, "defense": 0, "health": 0}
        bonus["speed"] = math.floor(self.fairy_exchanges / 10)

        for i in range(self.fairy_exchanges):
            bonus["strength"] += 2 if (i + 1) % 5 == 0 else 1
            bonus["defense"] += 2 if (i + 1) % 5 == 0 else 1
            bonus["health"] += 3 + math.floor(i / 2)
        return bonus


class Profile(BaseModel):
    """Profile.

    Args:
        profile_id (str): Id of profile
        cute_name (Optional[str]): Cute name of profile
        members (Dict[str, Members]): Dict of all members in profile.
    """

    profile_id: uuid.UUID
    cute_name: Optional[str]
    members: Dict[str, Members]
    banking: Optional[Dict[str, Any]]
    community_upgrades: Optional[Dict[str, Any]]



"""Watchdog objects."""
from pydantic import BaseModel
from pydantic import Field

from .utils import to_camel


class WatchDog(BaseModel):
    """Base class for watchdog.

    Args:
        watchdog_last_minute (int): Watchdog bans in last minute.
        staff_rolling_daily (int): Staff bans in the day.
        watchdog_total (int): Watchdog total bans.
        watchdog_rolling_daily (int): Watchdog bans in the day.
        staff_total (int): Staff total bans.
    """

    watchdog_last_minute: int = Field(alias="watchdog_lastMinute")
    staff_rolling_daily: int = Field(alias="staff_rollingDaily")
    watchdog_total: int = Field(alias="watchdog_total")
    watchdog_rolling_daily: int = Field(alias="watchdog_rollingDaily")
    staff_total: int = Field(alias="staff_total")

    class Config:
        """Config."""

        alias_generator = to_camel



"""Battleground."""
from pydantic import BaseModel


class Battleground(BaseModel):
    """Battleground games stats.

    Args:
        coins (int): Number of coins gathered in this Game Mode. Defaults to 0.
    """

    coins: int = 0



"""Build Battle."""
from pydantic import BaseModel


class BuildBattle(BaseModel):
    """Build Battle games stats.

    Args:
        coins (int): Number of coins gathered in this Game Mode. Defaults to 0.
        games_played (int): Total games played. Defaults to 0.
        score (int): Score. Defaults to 0.
    """

    coins: int = 0
    games_played: int = 0
    score: int = 0



"""Paintball."""
from pydantic import BaseModel


class Paintball(BaseModel):
    """Paintball games stats.

    Args:
        coins (int): Number of coins gathered in this Game Mode. Defaults to 0.
        deaths (int): Total deaths. Defaults to 0.
        wins (int): Total wins. Defaults to 0.
        shots_fired (int): Total shots fired. Defaults to 0.
    """

    coins: int = 0
    deaths: int = 0
    wins: int = 0
    shots_fired: int = 0



"""Status data class."""
from typing import Optional

from asyncpixel.constants import get_game_types
from pydantic import BaseModel
from pydantic import validator

from .game_type import GameType
from .utils import to_camel


class Status(BaseModel):
    """Status data object.

    Args:
        online (bool): Wether player is online.
        game_type (Optional[GameType]): Current game player is playing.
            Defaults to None.
        mode (Optional[str]): Mode of current game. Defaults to None.
    """

    online: bool = False
    game_type: Optional[GameType] = None
    mode: Optional[str] = None

    @validator("game_type", pre=True)
    @classmethod
    def validate_game_type(cls, v: str) -> GameType:
        """Validate game type."""
        game_type = [game for game in get_game_types() if game.type_name == v][0]
        return game_type

    class Config:
        """Config."""

        alias_generator = to_camel



"""Custom exceptions for asyncpixel."""
from .exceptions import ApiNoSuccessError
from .exceptions import InvalidApiKeyError
from .exceptions import RateLimitError

__all__ = ["ApiNoSuccessError", "InvalidApiKeyError", "RateLimitError"]



"""Game Count related objects."""
from typing import Dict
from typing import Optional

from pydantic import BaseModel

from .utils import to_camel


class GameCountsGame(BaseModel):
    """Game count game.

    Args:
        players (int): Number of players in a game.
        uuid_sender (Optional[Dict[str, int]]): Dict of game modes and people in them.
    """

    players: int
    modes: Optional[Dict[str, int]]


class GameCounts(BaseModel):
    """Game Count class.

    Args:
        games (Dict[str, GameCountsGame]): dict of all game and their game counts.
        player_count (int): total number of players online.
    """

    games: Dict[str, GameCountsGame]
    player_count: int

    class Config:
        """Config."""

        alias_generator = to_camel



"""Utils for pydantic models."""
from typing import Union


def safe_divide(dividend: Union[int, float], divisor: Union[int, float]) -> float:
    """Return dividend / divisor without raising ZeroDivisionError.

    Args:
        dividend (int, float): the dividend (a in a/b)
        divisor (int, float): the divisor (b in a/b)

    Returns:
        float: dividend / divisor, or floating point infinity if divisor == 0
    """
    if divisor == 0:
        return float("inf") * (1 if dividend >= 0 else -1)
    return dividend / divisor


def to_camel(string: str) -> str:
    """Convert str to camel case.

    Args:
        string (str): input.

    Returns:
        str: camelCase version of str.
    """
    out = string.split("_")[0]
    for word in string.split("_")[1:]:
        out += word.capitalize()
    return out



"""Vampirez."""
from pydantic import BaseModel


class VampireZ(BaseModel):
    """VampireZ games stats.

    Args:
        coins (int): Number of coins gathered in this Game Mode. Defaults to 0.
        human_deaths (int): Total deaths as a human. Defaults to 0.
        vampire_deaths (int): Total deaths as a vampire. Defaults to 0.
        zombie_kills (int): Total kills of a zombie. Defaults to 0.
        vampire_kills (int): Total kills of a vampire. Defaults to 0.
        most_vampire_kills_new (int): Most Vampire kills Defaults to 0.
    """

    coins: int = 0
    human_deaths: int = 0
    vampire_deaths: int = 0
    zombie_kills: int = 0
    vampire_kills: int = 0
    most_vampire_kills_new: int = 0



"""Housing."""
from pydantic import BaseModel


class Housing(BaseModel):
    """Housing games stats.

    Args:
        coins (int): Number of coins gathered in this Game Mode. Defaults to 0.
        total_kills (int): Total kills. Defaults to 0.
        total_wins (int): Total wins. Defaults to 0.
    """

    coins: int = 0
    total_kills: int = 0
    total_wins: int = 0



"""SkyClash."""
from pydantic import BaseModel


class SkyClash(BaseModel):
    """SkyClash games stats.

    Args:
        card_packs (int): Card packs. Defaults to 0.
    """

    card_packs: int = 0



"""Data objects for boosters."""
import datetime
import uuid
from typing import List
from typing import Union

from asyncpixel.constants import get_game_types
from pydantic import BaseModel
from pydantic import Field
from pydantic import validator

from .game_type import GameType
from .utils import to_camel


class Booster(BaseModel):
    """Main booster class.

    Args:
        id (str): ID
        purchaser_uuid (uuid.UUID): UUID of booster.
        amount (int): Amount of boosters.
        original_length (int): Original length of booster.
        length (int): Length of booster.
        game_type (GameType): Game type.
        date_activated (datetime.datetime): Date boost activated.
        stacked (Union[List[uuid.UUID], bool]): Wether boosters stacked.
    """

    id: str = Field(alias="_id")
    purchaser_uuid: uuid.UUID
    amount: float
    original_length: int
    length: int
    game_type: GameType

    @validator("game_type", pre=True)
    @classmethod
    def validate_game_type(cls, v: int) -> GameType:
        """Validate game type."""
        game_type = [game for game in get_game_types() if game.id == v][0]
        return game_type

    date_activated: datetime.datetime
    stacked: Union[List[uuid.UUID], bool] = False

    class Config:
        """Config."""

        alias_generator = to_camel


class Boosters(BaseModel):
    """Object containing boosters.

    Args:
        booster_state_decrementing (bool): Wether booster state decrementing.
        boosters (List[Booster]): List of boosters online.
    """

    booster_state_decrementing: bool = Field(alias="decrementing")
    boosters: List[Booster]



"""Asyncpixel.

An asyncronous hypixel api wrapper
"""
#  Copyright (C) 2020 Leon Bowie
# This program is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License a
# long with this program. If not, see <https://www.gnu.org/licenses/>.

try:
    from importlib.metadata import version, PackageNotFoundError  # type: ignore
except ImportError:  # pragma: no cover
    from importlib_metadata import version, PackageNotFoundError  # type: ignore

from .hypixel import Hypixel
from . import utils

try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
__author__ = "Leon Bowie"
__copyright__ = "Copyright 2020-2021 Leon Bowie"
__title__ = "asyncpixel"
__license__ = """Copyright (C) 2020 Leon Bowie

This program is free software: you can redistribute it
and/or modify it under the terms of the GNU General Public
License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License a
long with this program. If not, see <https://www.gnu.org/licenses/>."""

__all__ = [
    "__version__",
    "__author__",
    "__license__",
    "Hypixel",
    "utils",
]



"""Main class for key data."""
import uuid

from pydantic import BaseModel

from .utils import to_camel


class Key(BaseModel):
    """Main class for key data.

    Args:
        key (uuid.UUID): key text.
        owner (uuid.UUID): uuid of owner.
        limit (int): Limit of toal queries.
        queries_in_past_min (int): Queries in the past minute.
        total_queries (int): Total queries using the key.
    """

    key: uuid.UUID
    owner: uuid.UUID
    limit: int
    queries_in_past_min: int
    total_queries: int

    class Config:
        """Config."""

        alias_generator = to_camel



"""Player objects."""
import datetime
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from asyncpixel import utils
from asyncpixel.constants import get_game_types
from pydantic import BaseModel
from pydantic import Field
from pydantic import validator
from pydantic.class_validators import root_validator
from pydantic.types import UUID4

from .game_type import GameType
from .players import Arcade
from .players import Arena
from .players import Battleground
from .players import Bedwars
from .players import BuildBattle
from .players import Duels
from .players import GingerBread
from .players import Housing
from .players import HungerGames
from .players import Legacy
from .players import MCGO
from .players import Paintball
from .players import Pit
from .players import Quake
from .players import SkyClash
from .players import Skywars
from .players import TNTGames
from .players import UHC
from .players import VampireZ
from .players import Walls
from .players import Walls3
from .utils import to_camel


class Stats(BaseModel):
    """Game Stats.

    Args:
        bedwars (Optional[Bedwars]): bedwars stats.
        arcade (Optional[Arcade]): Arcade stats.
        build_battle (Optional[BuildBattle]): Build Battle stats.
        duels (Optional[Duels]): Duels stats.
        battleground (Optional[Battleground]): Battleground stats.
        hunger_games (Optional[HungerGames]): Hunger Games stats.
        ginger_bread (Optional[GingerBread]): Ginger Bread stats.
        paintball (Optional[Paintball]): Paintball stats.
        quake (Optional[Quake]): Quake stats.
        vampirez (Optional[VampireZ]): VampireZ stats.
        tnt_games (Optional[TNTGames]): TNT Games stats.
        uhc (Optional[UHC]): UHC stats.
        mcgo (Optional[MCGO]): MCGO stats.
        walls3 (Optional[Walls3]): Walls3 stats.
        walls (Optional[Walls]): Walls stats.
        arena (Optional[Arena]): Arena stats.
        sky_clash (Optional[SkyClash]): SkyClash stats.
        pit (Optional[Pit]): Pit stats.
        housing (Optional[Housing]): Housing stats.
        legacy (Optional[Legacy]): Legacy stats.
    """

    bedwars: Optional[Bedwars] = Field(alias="Bedwars")
    arcade: Optional[Arcade] = Field(alias="Arcade")
    build_battle: Optional[BuildBattle] = Field(alias="BuildBattle")
    duels: Optional[Duels] = Field(alias="Duels")
    battleground: Optional[Battleground] = Field(alias="Battleground")
    hunger_games: Optional[HungerGames] = Field(alias="HungerGames")
    ginger_bread: Optional[GingerBread] = Field(alias="GingerBread")
    paintball: Optional[Paintball] = Field(alias="Paintball")
    quake: Optional[Quake] = Field(alias="Quake")
    vampirez: Optional[VampireZ] = Field(alias="VampireZ")
    tnt_games: Optional[TNTGames] = Field(alias="TNTGames")
    uhc: Optional[UHC] = Field(alias="UHC")
    mcgo: Optional[MCGO] = Field(alias="MCGO")
    walls3: Optional[Walls3] = Field(alias="Walls3")
    walls: Optional[Walls] = Field(alias="Walls")
    arena: Optional[Arena] = Field(alias="Arena")
    sky_clash: Optional[SkyClash] = Field(alias="SkyClash")
    pit: Optional[Pit] = Field(alias="Pit")
    housing: Optional[Housing] = Field(alias="Housing")
    legacy: Optional[Legacy] = Field(alias="Legacy")
    skywars: Optional[Skywars] = Field(alias="SkyWars")

    # true_combat: Optional[TrueCombat] = Field(alias="TrueCombat")
    # speed_uhc: Optional[SpeedUHC]= Field(alias="SpeedUHC")
    # sky_block: Optional[SkyBlock]= Field(alias="SkyBlock")
    # super_smash: Optional[SuperSmash] = Field(alias="SuperSmash")


class Social(BaseModel):
    """Social accounts.

    Args:
        twitter (Optional[str]): Twitter.
        youtube (Optional[str]): YouTube.
        instagram (Optional[str]): Instagram.
        twitch (Optional[str]): Twitch.
        discord (Optional[str]): Discord.
        hypixel_forums (Optional[str]): Hypixel Forums.
    """

    twitter: Optional[str]
    youtube: Optional[str]
    instagram: Optional[str]
    twitch: Optional[str]
    discord: Optional[str]
    hypixel_forums: Optional[str]

    @root_validator(pre=True)
    @classmethod
    def get_social_media(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Get social media in correct format."""
        out = values.copy()
        for k, v in out["links"].items():
            out[k.lower()] = v
        return out


class Player(BaseModel):
    """Player.

    Args:
        uuid (UUID4): uuid of user.
        displayname (Optional[str]): Display name of user.
        rank (Optional[str]): Rank of user
        first_login (datetime.datetime): First login date.
        last_login (Optional[datetime.datetime]): Most recent login date.
        last_logout (Optional[datetime.datetime]): Last logout.
        stats (Stats): Stats for various game types.
        social_media (Optional[Social]): Social media accounts.
        id (Optional[str]): id of user.
        playername (Optional[str]): Playername.
        known_aliases (Optional[List[str]]): known aliases.
        known_aliases_lower (Optional[List[str]]): known aliases in lowercase.
        achievements_one_time (Optional[List[str]]): Achievements.
        mc_version_rp (Optional[str]): Minecraft version.
        network_exp (Optional[float]): Network experience.
        karma (Optional[int]): Player karma.
        last_adsense_generate_time (Optional[datetime.datetime]): Last generate
            time for adsense.
        last_claimed_reward (Optional[int]): Last claimed reward.
        total_rewards (Optional[int]): Total rewards.
        total_daily_rewards (Optional[int]): Total daily awards.
        reward_streak (Optional[int]): Current reward streak.
        reward_score (Optional[int]): Reward score.
        reward_high_score (Optional[int]): High score for rewards.
        friend_requests_uuid (Optional[List[UUID4]]): UUID of friend requests.
        achievement_tracking (Optional[List[str]]): Achievement tracking.
        achievement_points (Optional[int]): achievement points.
        current_gadget (Optional[str]): Current equipped gadget.
        channel (Optional[str]): Channel.
        most_recent_game_type (Optional[GameType]): Most recent Game Type that
            has been played.
        level (Optional[float]): Level of user.
        raw (Dict[str, Any]): raw data
    """

    uuid: UUID4
    displayname: Optional[str]
    rank: Optional[str]

    @root_validator(pre=True)
    @classmethod
    def create_rank(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Get the rank."""
        out = values.copy()
        rank = utils.get_rank(
            values.get("rank"),
            values.get("prefix"),
            values.get("monthlyPackageRank"),
            values.get("newPackageRank"),
            values.get("packageRank"),
        )
        out["rank"] = rank
        return out

    first_login: datetime.datetime
    last_login: Optional[datetime.datetime]
    last_logout: Optional[datetime.datetime]
    stats: Stats
    social_media: Optional[Social]

    id: Optional[str] = Field(alias="_id")
    playername: Optional[str]
    known_aliases: Optional[List[str]]
    known_aliases_lower: Optional[List[str]]
    achievements_one_time: Optional[List[str]]
    mc_version_rp: Optional[str]
    network_exp: Optional[float]
    karma: Optional[int]
    last_adsense_generate_time: Optional[datetime.datetime]
    last_claimed_reward: Optional[int]
    total_rewards: Optional[int]
    total_daily_rewards: Optional[int]
    reward_streak: Optional[int]
    reward_score: Optional[int]
    reward_high_score: Optional[int]
    friend_requests_uuid: Optional[List[UUID4]]
    achievement_tracking: Optional[List[str]]
    achievement_points: Optional[int]
    current_gadget: Optional[str]
    channel: Optional[str]
    most_recent_game_type: Optional[GameType]

    @validator("most_recent_game_type", pre=True)
    @classmethod
    def validate_game_type(cls, v: Union[str, int]) -> GameType:
        """Validate game type."""
        try:
            game_type = [game for game in get_game_types() if game.id == v][0]
        except Exception:
            game_type = [game for game in get_game_types() if game.type_name == v][0]
        return game_type

    level: float

    @root_validator(pre=True)
    @classmethod
    def create_level(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Create level."""
        out = values.copy()
        exp = float(out.get("networkExp", 0.0))
        out["level"] = utils.calc_player_level(exp)
        return out

    raw: Dict[str, Any]

    @root_validator(pre=True)
    @classmethod
    def create_raw(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Create a copy of the values to directly access."""
        out = values.copy()
        out["raw"] = out
        return out

    class Config:
        """Config."""

        alias_generator = to_camel



"""Walls."""
from pydantic import BaseModel


class Walls(BaseModel):
    """Walls games stats.

    Args:
        coins (int): Number of coins gathered in this Game Mode. Defaults to 0.
        losses (int): Games lost. Defaults to 0.
        deaths (int): Total deaths. Defaults to 0.
    """

    coins: int = 0
    losses: int = 0
    deaths: int = 0



"""Player Stats."""
from .arcade import Arcade
from .arena import Arena
from .battleground import Battleground
from .bedwars import Bedwars
from .bedwars import BedwarsGame
from .build_battle import BuildBattle
from .duels import Duels
from .gingerbread import GingerBread
from .housing import Housing
from .hunger_games import HungerGames
from .legacy import Legacy
from .mcgo import MCGO
from .paintball import Paintball
from .pit import Pit
from .quake import Quake
from .skyclash import SkyClash
from .skywars import Skywars
from .tnt_games import TNTGames
from .uhc import UHC
from .vampirez import VampireZ
from .walls import Walls
from .walls3 import Walls3

__all__ = [
    "Arcade",
    "Arena",
    "Battleground",
    "Bedwars",
    "BuildBattle",
    "Duels",
    "GingerBread",
    "Housing",
    "HungerGames",
    "Legacy",
    "MCGO",
    "Paintball",
    "Pit",
    "Quake",
    "SkyClash",
    "TNTGames",
    "UHC",
    "VampireZ",
    "Walls",
    "Walls3",
    "BedwarsGame",
    "Skywars",
]



"""Custom Exceptions for asyncpixel."""
import datetime


class RateLimitError(Exception):
    """Exception raised when Hypixel ratelimit is reached."""

    def __init__(self, retry_after: datetime.datetime) -> None:
        """Create error."""
        self.message = (
            "The hypixel API ratelimit was reached, "
            + f"try again at {retry_after.strftime('%H:%M:%S')}."
        )
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return error in readable format.

        Returns:
            str: string version of error.
        """
        return self.message


class ApiNoSuccessError(Exception):  # noqa: N8181
    """Exception raised when api has an error."""

    def __init__(
        self,
        source: str,
    ) -> None:
        """Create error."""
        self.message = f"The {source} endpoint encounted an error on the hypixel side."
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return error in readable format.

        Returns:
            str: string version of error.
        """
        return self.message


class InvalidApiKeyError(Exception):
    """Exception raised when the API key is invalid."""

    def __init__(self, message: str = "Entered API key is not valid") -> None:
        """API key not valid..

        Args:
            message (str): error message. Defaults to "Entered API key is not valid".
        """
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return error in readable format.

        Returns:
            str: string version of error.
        """
        return self.message



"""Game type."""
from typing import Optional

from pydantic import BaseModel


class GameType(BaseModel):
    """Main game class.

    Args:
        id (str): ID
        purchaser_uuid (uuid.UUID): UUID of booster.
        amount (int): Amount of boosters.
        original_length (int): Original length of booster.
        length (int): Length of booster.
        game_type (int): Game type.
        date_activated (datetime.datetime): Date boost activated.
        stacked (Union[List[uuid.UUID], bool]): Wether boosters stacked.
    """

    id: int
    type_name: str
    database_name: str
    lobby_name: Optional[str]
    clean_name: str
    standard_name: str
    legacy: bool = False



"""Game Count related objects."""
import uuid
from typing import List
from typing import Tuple

from pydantic import BaseModel


class Leaderboards(BaseModel):
    """Game count game.

    Args:
        path (int): Path.
        prefix (str): Prefix.
        title (str): Title of leaderboard.
        location (Tuple[int, int, int]): Location of leaderboard in lobby.
        count (int): Count of leaderboards.
        leaders (List[uuid.UUID]): List of leaders on the leaderboard.
    """

    path: str
    prefix: str
    title: str
    location: Tuple[int, int, int]
    count: int
    leaders: List[uuid.UUID]



"""Models for asyncpixel data objects."""
from .auctions import Auction
from .auctions import AuctionItem
from .auctions import Bids
from .bazaar import Bazaar
from .bazaar import BazaarItem
from .bazaar import BazaarQuickStatus
from .bazaar import BazaarSummary
from .booster import Booster
from .booster import Boosters
from .friends import Friend
from .game_count import GameCounts
from .game_count import GameCountsGame
from .game_type import GameType
from .games import Game
from .guild import Guild
from .guild import GuildMembers
from .guild import Pattern
from .guild import Rank
from .key import Key
from .leaderboards import Leaderboards
from .news import Item
from .news import News
from .player import Player
from .player import Social
from .player import Stats
from .profile import InvArmor
from .profile import Members
from .profile import Objective
from .profile import Profile
from .profile import Quests
from .status import Status
from .watchdog import WatchDog

__all__ = [
    "Auction",
    "Item",
    "AuctionItem",
    "Bazaar",
    "BazaarSummary",
    "BazaarItem",
    "BazaarQuickStatus",
    "Booster",
    "Boosters",
    "Friend",
    "GameCounts",
    "GameCountsGame",
    "Game",
    "Guild",
    "Key",
    "Leaderboards",
    "News",
    "Player",
    "InvArmor",
    "Members",
    "Objective",
    "Profile",
    "Quests",
    "Status",
    "WatchDog",
    "Bids",
    "GuildMembers",
    "Banner",
    "Pattern",
    "GameType",
    "Stats",
    "Social",
    "Rank",
]



"""Game related objects."""
import datetime
from typing import Optional

from asyncpixel.constants import get_game_types
from pydantic import BaseModel
from pydantic import validator

from .game_type import GameType
from .utils import to_camel


class Game(BaseModel):
    """Game class.

    Args:
        date (datetime.datetime): Time game started.
        game_type (GameType): Game Type.
        mode (Optional[str]): Game mode.
        map (Optional[str]): Map the game is on.
        ended (Optional[datetime.datetime]): Time game ended. Defaults to None.
    """

    date: datetime.datetime
    game_type: GameType
    mode: Optional[str]
    map: Optional[str]
    ended: Optional[datetime.datetime] = None

    @validator("game_type", pre=True)
    @classmethod
    def validate_game_type(cls, v: str) -> GameType:
        """Turn game type to correct format."""
        game_type = [game for game in get_game_types() if game.type_name == v][0]
        return game_type

    class Config:
        """Config."""

        alias_generator = to_camel



"""Utils."""
import re
from typing import Optional
from typing import Union

ranks = {
    "NONE": None,
    "VIP": "VIP",
    "VIP_PLUS": "VIP+",
    "MVP": "MVP",
    "MVP_PLUS": "MVP+",
    "SUPERSTAR": "MVP++",
    "YOUTUBER": "YOUTUBE",
    "PIG+++": "PIG+++",
    "BUILD TEAM": "BUILD TEAM",
    "HELPER": "HELPER",
    "MODERATOR": "MOD",
    "ADMIN": "ADMIN",
    "SLOTH": "SLOTH",
    "OWNER": "OWNER",
}


def get_rank(
    package_rank: Optional[str] = None,
    rank: Optional[str] = None,
    prefix_raw: Optional[str] = None,
    monthly_package_rank: Optional[str] = None,
    new_package_rank: Optional[str] = None,
) -> Optional[str]:
    """Get rank of Hypixel player.

    Args:
        rank (Optional[str]): rank
        prefix_raw (Optional[str]): raw prefix
        monthly_package_rank (Optional[str]): monthly package for mvp++
        new_package_rank (Optional[str]): new rank format
        package_rank (Optional[str]): old rank format

    Returns:
        Optional[str]: rank
    """
    real_rank = None
    if prefix_raw:
        prefix = re.sub(r"§.", "", prefix_raw)[1:-1]
        # prefixes all start and end with brackets, and have minecraft color codes,
        # this is to remove color codes and
        # brackets
        real_rank = ranks.get(prefix, prefix)
    elif rank and rank != "NORMAL" and not real_rank:
        real_rank = ranks.get(rank, rank)
    elif (monthly_package_rank and monthly_package_rank != "NONE") and not real_rank:
        real_rank = ranks.get(monthly_package_rank, monthly_package_rank)
    elif new_package_rank and not real_rank:
        real_rank = ranks.get(new_package_rank, new_package_rank)
    elif package_rank and not real_rank:
        real_rank = ranks.get(package_rank, package_rank)
    return real_rank


def calc_player_level(xp: Union[float, int]) -> float:
    """Calculate player level from xp.

    Args:
        xp (int): amount of xp a player has.

    Returns:
        float: current level of player.
    """
    return float(1 + (-8750.0 + (8750**2 + 5000 * xp) ** 0.5) / 2500)



"""Bazaar classes."""
import datetime
from typing import List

from pydantic import BaseModel

from .utils import to_camel


class BazaarSummary(BaseModel):
    """Bazaar object.

    Args:
        amount (int): Amount available to buy.
        price_per_unit (float): Price per unit.
        orders (int): How many orders.
    """

    amount: int
    price_per_unit: float
    orders: int

    class Config:
        """Config."""

        alias_generator = to_camel


class BazaarQuickStatus(BaseModel):
    """Bazaar quick status.

    Args:
        product_id (str): Id of product.
        sell_price (int): Sell price per unit.
        sell_volume (float): Volume of sale.
        sell_moving_week (int): Sell moving week.
        sell_orders (int): How many orders.
        buy_price (int): Buy price.
        buy_volume (float): Volume of purchase.
        buy_moving_week (int): How many orders.
        buy_orders (int): How many orders.
    """

    product_id: str
    sell_price: float
    sell_volume: int
    sell_moving_week: int
    sell_orders: int
    buy_price: float
    buy_volume: int
    buy_moving_week: int
    buy_orders: int

    class Config:
        """Config."""

        alias_generator = to_camel


class BazaarItem(BaseModel):
    """Bazaar item.

    Args:
        product_id (str): Product id.
        buy_summary (List[BazaarSellSummary]): List of sell summary.
        buy_summary (List[BazaarBuySummary]): List of buy summary.
        quick_status (BazaarQuickStatus): Quick status.
    """

    product_id: str
    sell_summary: List[BazaarSummary]
    buy_summary: List[BazaarSummary]
    quick_status: BazaarQuickStatus


class Bazaar(BaseModel):
    """Bazaar object.

    Args:
        last_updated (datetime.datetime): Time last updated.
        bazaar_items (List[BazaarItem]): Items in bazaar.
    """

    last_updated: datetime.datetime
    bazaar_items: List[BazaarItem]



"""MCGO."""
from pydantic import BaseModel
from pydantic import Field


class MCGO(BaseModel):
    """MCGO games stats.

    Args:
        coins (int): Number of coins gathered in this Game Mode. Defaults to 0.
        deaths (int): Total deaths. Defaults to 0.
        shots_fired (int): Total shots fired. Defaults to 0.
        round_wins (int): Total round wins. Defaults to 0.
        bombs_planted (int): Total bombs planted. Defaults to 0.
        game_wins_deathmatch (int): Total games won of deathmatch. Defaults to 0.
        wins (int): Total wins. Defaults to 0.
        kills (int): Total kills. Defaults to 0.
    """

    coins: int = 0
    deaths: int = 0
    shots_fired: int = 0
    round_wins = 0
    bombs_planted: int = 0
    game_wins_deathmatch: int = 0
    wins: int = Field(0, alias="game_wins")
    kills: int = 0



"""News related objects."""
from typing import Optional

from pydantic import BaseModel


class Item(BaseModel):
    """News Item.

    Args:
        material (str): Material of article.
        data (Optional[int]): Data.
    """

    material: str
    data: Optional[int]


class News(BaseModel):
    """News object.

    Args:
        item (Item): News item.
        link (str): Link to article.
        text (text): Text of news.
        title (str): Title of news article.
    """

    item: Item
    link: str
    text: str
    title: str



"""Auction related objects."""
import datetime
from math import floor
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from pydantic import BaseModel
from pydantic import Field
from pydantic.types import UUID4

from .utils import to_camel


class Bids(BaseModel):
    """Bid models.

    Args:
        auction_id (UUID4): Id of auction.
        bidder (UUID4): Id of bidder.
        profile_id (UUID4): Profile_id of seller.
        amount (str): Amount bidded.
        timestamp (datetime.datetime): Timestamp of bid placed.
    """

    auction_id: UUID4
    bidder: UUID4
    profile_id: Optional[UUID4]
    amount: int
    timestamp: datetime.datetime


class AuctionItem(BaseModel):
    """Auction model.

    Args:
        uuid (UUID4): Id of auction.
        auctioneer (UUID4): Id of seller.
        profile_id (UUID4): Profile_id of seller.
        coop (str): Amount bidded.
        start (datetime.datetime): Start time of auction.
        end (datetime.datetime): End time of auction.
        item_name (str): Name of auction item.
        item_lore (str): Lore of item.
        extra (str): extra.
        category (str): Item category
        tier (str): Tier of item.
        starting_bid (int): Starting Auction bid.
        item_bytes (str): Bytes of item.
        claimed (bool): Wehther the auction has been won.
        claimed_bidders (Optional[List[UUID4]]): Amount bidded.
        highest_bid_amount (int): Highest amount bidded.
        bids (List[Bids]): List of bids on auction.
        id (str): Id of auction.
        bin (bool): Whether the auction is BIN (Buy instantly)

    """

    uuid: UUID4
    auctioneer: UUID4
    profile_id: UUID4
    coop: List[UUID4]
    start: datetime.datetime
    end: datetime.datetime
    item_name: str
    item_lore: str
    extra: str
    category: str
    tier: str
    starting_bid: int
    item_bytes: Union[str, Dict[str, Union[int, str]]]
    claimed: bool
    claimed_bidders: Optional[List[UUID4]]
    highest_bid_amount: int
    bids: List[Bids]
    id: Optional[str] = Field(alias="_id")
    bin: bool = False

    def active(self) -> bool:
        """Return if auction is active - you can bid on it."""
        return not self.claimed and datetime.datetime.now(tz=self.end.tzinfo) < self.end

    def lowest_possible_bid(self) -> int:
        """Returns next lowest possible bid."""
        current_price = max(self.starting_bid, self.highest_bid_amount)

        if 1 <= current_price <= 3:
            return floor(current_price) + 1

        return floor(current_price * 1.15)


class Auction(BaseModel):
    """Main auction object.

    Args:
        page (int): Page of auction data.
        total_pages (int): Total pages of auctions.
        total_auctions (int): Total number of auctions.
        last_updated (datetime.datetime): Time last updated.
        auctions (List[AuctionItem]): List of auctions.
    """

    page: int
    total_pages: int
    total_auctions: int
    last_updated: datetime.datetime
    auctions: List[AuctionItem]

    class Config:
        """Config."""

        alias_generator = to_camel



"""Duels."""
from pydantic import BaseModel


class Duels(BaseModel):
    """Duels games stats.

    Args:
        coins (int): Number of coins gathered in this Game Mode. Defaults to 0.
        deaths (int): Totaldeaths. Defaults to 0.
        losses (int): Games lost. Defaults to 0.
        losses (int): Games won. Defaults to 0.
        damage_dealt (int): Damge dealt in total. Defaults to 0.
        rounds_played (int): Number of rounds played. Defaults to 0.
    """

    coins: int = 0
    deaths: int = 0
    losses: int = 0
    wins: int = 0
    damage_dealt: int = 0
    rounds_played: int = 0



"""Bedwars."""
import warnings
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from asyncpixel.models.utils import safe_divide
from pydantic import BaseModel
from pydantic import Field
from pydantic import root_validator

# Amount of levels to prestige
LEVELS_PER_PRESTIGE = 100

# The exp required to level up once
LEVEL_COST = 5000

# The exp required to level up to the first few levels after a prestige
EASY_LEVEL_COSTS = {1: 500, 2: 1000, 3: 2000, 4: 3500}

# The exp required to level up past the easy levels
EASY_EXP = sum(EASY_LEVEL_COSTS.values())

# The amount of easy levels
EASY_LEVELS = len(EASY_LEVEL_COSTS)

# The exp required to prestige
PRESTIGE_EXP = EASY_EXP + (100 - EASY_LEVELS) * LEVEL_COST


def bedwars_level_from_exp(exp: int) -> float:
    """Bedwars level/star.

    Args:
        exp (int): the player's bedwars experience

    Returns:
        float: bedwars level + progress towards next level
    """
    levels = (exp // PRESTIGE_EXP) * LEVELS_PER_PRESTIGE
    exp %= PRESTIGE_EXP

    # The first few levels have different costs
    for i in range(1, EASY_LEVELS + 1):
        cost = EASY_LEVEL_COSTS[i]
        if exp >= cost:
            levels += 1
            exp -= cost
        else:
            # We can't afford the next level, so we have found the level we are at
            break

    levels += exp // LEVEL_COST
    exp %= LEVEL_COST

    next_level = (levels + 1) % LEVELS_PER_PRESTIGE

    # The cost of the next level
    if next_level in EASY_LEVEL_COSTS:
        next_level_cost = EASY_LEVEL_COSTS[next_level]
    else:
        next_level_cost = LEVEL_COST

    return levels + exp / next_level_cost


class BedwarsGame(BaseModel):
    """Bedwars GameMode stats.

    Args:
        items_purchased (int): Total Items purchased. Defaults to 0.
        diamond_resources_collected (int): Total diamonds collected. Defaults to 0.
        games_played (int): Number of games played. Defaults to 0.
        losses (int): Games lost for this Game Mode. Defaults to 0.
        gold_resources_collected (int): Total gold collected. Defaults to 0.
        void_deaths (int): Deaths to the void Defaults to 0.
        deaths (int): Total deaths. Defaults to 0.
        winstreak (int): Current winstreak. Defaults to 0.
        beds_lost (int): Total beds lost. Defaults to 0.
        final_deaths (int): Total final deaths. Defaults to 0.
        entity_attack_deaths (int): Number of deaths to an entity. Defaults to 0.
        beds_broken (int): Beds broken. Defaults to 0.
        entity_attack_final_deaths (int): Number of deaths to an entity. Defaults to 0.
        fall_deaths (int): Total deaths from fall damage. Defaults to 0.
        magic_deaths (int): Total deaths to magic. Defaults to 0.
        permanent_items_purchased (int): Number of items that are not lost
            in the game purchased. Defaults to 0.
        void_kills (int): Kills using the void. Defaults to 0.
        kills (int): Total Total kills. Defaults to 0.
        wins (int): Total games won. Defaults to 0.
        void_final_deaths (int): Total final deaths to the void. Defaults to 0.
        final_kills (int): Total final kills. Defaults to 0.
    """

    items_purchased: int = Field(0, alias="items_purchased_bedwars")
    diamond_resources_collected: int = Field(
        0, alias="diamond_resources_collected_bedwars"
    )
    games_played: int = Field(0, alias="games_played_bedwars")
    losses: int = Field(0, alias="losses_bedwars")
    iron_resources_collected: int = Field(0, alias="iron_resources_collected_bedwars")
    resources_collected: int = Field(0, alias="resources_collected_bedwars")
    gold_resources_collected: int = Field(0, alias="gold_resources_collected_bedwars")
    void_deaths: int = Field(0, alias="void_deaths_bedwars")
    deaths: int = Field(0, alias="deaths_bedwars")
    winstreak: int = Field(0, alias="winstreak")
    beds_lost: int = Field(0, alias="beds_lost_bedwars")
    final_deaths: int = Field(0, alias="final_deaths_bedwars")
    entity_attack_deaths: int = Field(0, alias="entity_attack_deaths_bedwars")
    beds_broken: int = Field(0, alias="beds_broken_bedwars")
    entity_attack_final_deaths: int = Field(
        0, alias="entity_attack_final_deaths_bedwars"
    )
    fall_deaths: int = Field(0, alias="fall_deaths_bedwars")
    magic_deaths: int = Field(0, alias="magic_deaths_bedwars")
    permanent_items_purchased: int = Field(
        0, alias="permanent _items_purchased_bedwars"
    )
    void_kills: int = Field(0, alias="void_kills_bedwars")
    kills: int = Field(0, alias="kills_bedwars")
    wins: int = Field(0, alias="wins_bedwars")
    void_final_deaths: int = Field(0, alias="void_final_deaths_bedwars")
    final_kills: int = Field(0, alias="final_kills_bedwars")

    @property
    def win_per_lose(self) -> float:
        """Wins per losses.

        Returns:
            float: ratio between game wins and game losses.
        """
        return safe_divide(self.wins, self.losses)

    @property
    def beds_broken_per_lost(self) -> float:
        """Beds broken per Beds lost.

        Returns:
            float: ratio between beds broken and lost.
        """
        return safe_divide(self.beds_broken, self.beds_lost)

    @property
    def final_kills_per_kills(self) -> float:
        """Final kills per normal kill.

        Returns:
            float: ratio between final kills and normal kills
        """
        warnings.warn(
            "Field final_kills_per_kills will be removed in a future update",
            DeprecationWarning,
        )
        return safe_divide(self.final_kills, self.kills)

    @property
    def final_kills_per_final_death(self) -> float:
        """Final kills per final death.

        Returns:
            float: ratio between final kills and final deaths.
        """
        return safe_divide(self.final_kills, self.final_deaths)


class Bedwars(BaseModel):
    """Bedwars Stats.

    Args:
        kills (int): Total kills across all Bedwars gamemodes. Defaults to 0.
        wins (int): Total wins across all Bedwars gamemodes. Defaults to 0.
        coins (int): Total coins collected. Defaults to 0.
        games_played (int): Number of bedwars games played. Defaults to 0.
        final_deaths (int): Number of final deaths across all Bedwars gamemodes.
            Defaults to 0.
        deaths (int): Number of deaths across all Bedwars gamemodes. Defaults to 0.
        final_kills (int): Number of final kills across all Bedwars gamemodes.
            Defaults to 0.
        losses (int): Total bedwars games lost. Defaults to 0.
        beds_lost (int): Number of beds lost across all Bedwars gamemodes.
            Defaults to 0.
        beds_broken (int): Number of beds broken across all Bedwars gamemodes.
            Defaults to 0.
        winstreak (int): Current winstreak. Defaults to 0.
        experience (int): Total bedwars experience. Defaults to 0.
        singles (Optional[BedwarsGame]): Stats for the singles Gamemode.
        doubles (Optional[BedwarsGame]): Stats for the doubles Gamemode.
        triples (Optional[BedwarsGame]): Stats for the triples Gamemode.
        quads (Optional[BedwarsGame]): Stats for the quads Gamemode.
        four_v_four (Optional[BedwarsGame]): Stats for the four vs four Gamemode.
        quads_ultimate (Optional[BedwarsGame]): Stats for the quads ultimate Gamemode.
        doubles_ultimate (Optional[BedwarsGame]): Stats for the doubles
            ultimate Gamemode.
        castle (Optional[BedwarsGame]): Stats for the castle Gamemode.
        entity_attack_deaths (int): Deaths to an entity. Defaults to 0.
        entity_attack_final_deaths (int): Final deaths to an entity. Defaults to 0.
        fall_deaths (int): Overall fall deaths. Defaults to 0.
        fall_final_deaths (int): Overall final fall deaths Defaults to 0.
        projectile_deaths (int): Projectile deaths. Defaults to 0.
        suffocation_deaths (int): Overall suffocation deaths. Defaults to 0.
        magic_deaths (int): Overall magic deaths. Defaults to 0.
        entity_explosion_deaths (int): Overall explosion deaths. Defaults to 0.
        magic_final_deaths (int): Overall magic final deaths. Defaults to 0.
        void_deaths (int): Overall void deaths. Defaults to 0.
        void_final_death (int): Overall void final deaths. Defaults to 0.
        fire_tick_final_death (int): Overall fire final deaths. Defaults to 0.
        void_kills (int): Overall void kills. Defaults to 0.
        entity_attack_kills (int): Overall entity attack kills. Defaults to 0.
        entity_attack_final_kills (int): Overall entity attack final kills.
            Defaults to 0.
        projectile_final_kills (int): Overall projectile final kills. Defaults to 0.
        entity_explosion_kills (int): Overall entity explosion kills. Defaults to 0.
        projectile_kills (int): Overall projectile kills. Defaults to 0.
        fall_final_kills (int): Overall final fall kills. Defaults to 0.
        fall_kills (int): Overall fall kills. Defaults to 0.
        resources_collected (int): Overall resources collected. Defaults to 0.
        iron_resources_collected (int): Overall iron resources collected. Defaults to 0.
        diamond_resources_collected (int): Overall diamond resources collected.
            Defaults to 0.
        emerald_resources_collected (int): Overall emerald resources collected.
            Defaults to 0.
        items_purchased (int): Overall items purchased. Defaults to 0.
        permanent_items_purchased (int): Overall permanent items purchased.
            Defaults to 0.
        bedwars_box_rares (int): Bedwars boxes that are rare. Defaults to 0.
        bedwars_box (int): Total Bedwars boxes. Defaults to 0.
        chest_history_new (List[str]): Chest history.  Defaults to [].
        bedwars_box_commons (int): Total common Bedwars boxes. Defaults to 0.
        spray_glyph_field (str): Current spray glyph.  Defaults to "".
        active_island_topper (str): Current island topper.  Defaults to "".
        active_projectile_trail (str): Current projectile trail.  Defaults to "".
        bedwars_easter_boxes (int): Bedwars easters boxes. Defaults to 0.
        Bedwars_opened_chests (int): Total chests opened in Bedwars. Defaults to 0.
        Bedwars_opened_rares (int): Total rare chests opened in Bedwars.
            Defaults to 0.
        Bedwars_opened_commons (int): Total common chests opened in Bedwars.
            Defaults to 0.
        active_npc_skin (str): Current active NPC skin.  Defaults to "".
        favourites_2 (str): Favourites.  Defaults to "".
        Bedwars_opened_epics (int): Total epic chests opened in Bedwars.
            Defaults to 0.
        active_death_cry (str): Equiped active death cry.  Defaults to "".
        active_kill_effect (str): Equiped kill effect.  Defaults to "".
        active_sprays (str): Equiped spray.  Defaults to "".
        active_glyph (str): Active glyph.  Defaults to "".
        selected_ultimate (str): Selected Ultimate.  Defaults to "".
    """

    kills: int = Field(0, alias="kills_bedwars")
    wins: int = Field(0, alias="wins_bedwars")
    coins: int = Field(0, alias="coins")
    games_played: int = Field(0, alias="games_played_bedwars")
    final_deaths: int = Field(0, alias="final_deaths_bedwars")
    deaths: int = Field(0, alias="deaths_bedwars")
    final_kills: int = Field(0, alias="final_kills_bedwars")
    losses: int = Field(0, alias="losses_bedwars")
    beds_lost: int = Field(0, alias="beds_lost_bedwars")
    beds_broken: int = Field(0, alias="beds_broken_bedwars")
    winstreak: int = Field(0, alias="winstreak")
    experience: int = Field(0, alias="Experience")

    singles: Optional[BedwarsGame]
    doubles: Optional[BedwarsGame]
    triples: Optional[BedwarsGame]
    quads: Optional[BedwarsGame]
    four_v_four: Optional[BedwarsGame]
    quads_ultimate: Optional[BedwarsGame]
    doubles_ultimate: Optional[BedwarsGame]
    castle: Optional[BedwarsGame]

    # Deaths
    entity_attack_deaths: int = Field(0, alias="entity_attack_deaths_bedwars")
    entity_attack_final_deaths: int = Field(
        0, alias="entity_attack_final_deaths_bedwars"
    )
    fall_deaths: int = Field(0, alias="fall_deaths_bedwars")
    fall_final_deaths: int = Field(0, alias="fall_final_deaths_bedwars")
    projectile_deaths: int = Field(0, alias="projectile_deaths_bedwars")
    suffocation_deaths: int = Field(0, alias="suffocation_deaths_bedwars")
    magic_deaths: int = Field(0, alias="magic_deaths_bedwars")
    entity_explosion_deaths: int = Field(0, alias="entity_explosion_deaths_bedwars")
    magic_final_deaths: int = Field(0, alias="magic_final_deaths_bedwars")
    void_deaths: int = Field(0, alias="void_deaths_bedwars")
    void_final_death: int = Field(0, alias="void_final_deaths_bedwars")
    fire_tick_final_death: int = Field(0, alias="fire_tick_final_deaths_bedwars")

    # Kills
    void_kills: int = Field(0, alias="void_kills_bedwars")
    entity_attack_kills: int = Field(0, alias="entity_attack_kills_bedwars")
    entity_attack_final_kills: int = Field(0, alias="entity_attack_final_kills_bedwars")
    projectile_final_kills: int = Field(0, alias="projectile_final_kills_bedwars")
    entity_explosion_kills: int = Field(0, alias="entity_explosion_kills_bedwars")
    projectile_kills: int = Field(0, alias="projectile_kills_bedwars")
    fall_final_kills: int = Field(0, alias="fall_final_kills_bedwars")
    fall_kills: int = Field(0, alias="fall_kills_bedwars")

    # Resources
    resources_collected: int = Field(0, alias="resources_collected_bedwars")
    iron_resources_collected: int = Field(0, alias="iron_resources_collected_bedwars")
    diamond_resources_collected: int = Field(
        0, alias="diamond_resources_collected_bedwars"
    )
    emerald_resources_collected: int = Field(
        0, alias="emerald_resources_collected_bedwars"
    )

    # Purchases
    items_purchased: int = Field(0, alias="_items_purchased_bedwars")
    permanent_items_purchased: int = Field(0, alias="permanent_items_purchased_bedwars")

    # Other
    bedwars_box_rares: int = Field(0, alias="bedwars_box_rares")
    bedwars_box: int = Field(0, alias="bedwars_box")
    chest_history_new: List[str] = Field([], alias="chest_history_new")
    bedwars_box_commons: int = Field(0, alias="bedwars_box_commons")
    spray_glyph_field: str = Field("", alias="spray_glyph_field")
    active_island_topper: str = Field("", alias="activeIslandTopper")
    active_projectile_trail: str = Field("", alias="activeProjectileTrail")
    bedwars_easter_boxes: int = Field(0, alias="beds_lost_bedwars")
    Bedwars_opened_chests: int = Field(0, alias="Bedwars_openedChests")
    Bedwars_opened_rares: int = Field(0, alias="Bedwars_openedRares")
    Bedwars_opened_commons: int = Field(0, alias="Bedwars_openedCommons")
    active_npc_skin: str = Field("", alias="activeNPCSkin")
    favourites_2: str = Field("", alias="favourites_2")
    Bedwars_opened_epics: int = Field(0, alias="Bedwars_openedEpics")
    active_death_cry: str = Field("", alias="activeDeathCry")
    active_kill_effect: str = Field("", alias="activeKillEffect")
    active_sprays: str = Field("", alias="activeSprays")
    active_glyph: str = Field("", alias="activeGlyph")
    selected_ultimate: str = Field("", alias="selected_ultimate")

    @property
    def final_kills_per_kills(self) -> float:
        """Final kills per normal kill.

        Returns:
            float: ratio between final kills and normal kills
        """
        warnings.warn(
            "Field final_kills_per_kills will be removed in a future update",
            DeprecationWarning,
        )
        return safe_divide(self.final_kills, self.kills)

    @property
    def final_kills_per_final_death(self) -> float:
        """Final kills per final death.

        Returns:
            float: ratio between final kills and final deaths.
        """
        return safe_divide(self.final_kills, self.final_deaths)

    @property
    def beds_broken_per_lost(self) -> float:
        """Beds broken per Beds lost.

        Returns:
            float: ratio between beds broken and beds lost.
        """
        return safe_divide(self.beds_broken, self.beds_lost)

    @property
    def win_lose(self) -> float:
        """Wins per losses.

        Returns:
            float: ratio between game wins and game losses.
        """
        return safe_divide(self.wins, self.losses)

    @property
    def level(self) -> float:
        """Bedwars level/star.

        Returns:
            float: bedwars level + progress towards next level
        """
        return bedwars_level_from_exp(self.experience)

    @root_validator(pre=True)
    @classmethod
    def traverse_sources(cls, values: Dict[str, Any]) -> Dict[str, Any]:  # noqa: C901
        """Traverse all sources."""
        out = values.copy()
        for (name, field) in values.items():
            if name.startswith("eight_one_"):
                if "singles" not in out:
                    out["singles"] = {}

                out["singles"][name[10:]] = field
            elif name.startswith("eight_two_"):
                if "doubles" not in out:
                    out["doubles"] = {}

                out["doubles"][name[10:]] = field
            elif name.startswith("four_three_"):
                if "triples" not in out:
                    out["triples"] = {}

                out["triples"][name[11:]] = field
            elif name.startswith("four_four_"):
                if "quads" not in out:
                    out["quads"] = {}

                out["quads"][name[10:]] = field
            elif name.startswith("two_four_"):
                if "four_v_four" not in out:
                    out["four_v_four"] = {}

                out["four_v_four"][name[9:]] = field
            elif name.startswith("quads_ultimate_"):  # pragma: no cover
                if "quads_ultimate" not in out:
                    out["quads_ultimate"] = {}

                out["quads_ultimate"][name[15:]] = field
            elif name.startswith("ultimate_"):  # pragma: no cover
                if "doubles_ultimate" not in out:
                    out["doubles_ultimate"] = {}

                out["doubles_ultimate"][name[9:]] = field
            elif name.startswith("castle_"):
                if "castle" not in out:
                    out["castle"] = {}

                out["castle"][name[7:]] = field

        return out



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['asyncpixel',
 'asyncpixel.exceptions',
 'asyncpixel.models',
 'asyncpixel.models.players']

package_data = \
{'': ['*'],
 'asyncpixel': ['hypixelconstants/*',
                'hypixelconstants/.github/workflows/*',
                'hypixelconstants/build/*',
                'hypixelconstants/json/*',
                'hypixelconstants/tasks/*']}

install_requires = \
['aiohttp>=3.6.2,<4.0.0', 'pydantic>=1.7.3,<2.0.0']

extras_require = \
{':python_version < "3.8"': ['importlib_metadata>=2,<5']}

setup_kwargs = {
    'name': 'asyncpixel',
    'version': '1.3.2',
    'description': 'Asyncpixel',
    'long_description': '<h1 align="center">\n  Asyncpixel\n</h1>\n\n<h3 align="center">\n  Easily access Hypixel\'s api\n</h3>\n<p align="center">\n  Asyncpixel is an open source asynchronous python wrapper for the Hypixel api with 100% of all endpoints.\n</p>\n\n<h3 align="center">\n 🤖 🎨 🚀\n</h3>\n\n<p align="center">\n  <a href="https://github.com/Darkflame72/asyncpixel/releases">\n    <img alt="GitHub all releases" src="https://img.shields.io/github/downloads/Darkflame72/asyncpixel/total">\n  </a>\n  <a href="https://github.com/Darkflame72/asyncpixel/releases">\n    <img alt="GitHub release (latest by date)" src="https://img.shields.io/github/v/release/Darkflame72/asyncpixel">\n  </a>\n  <a href="https://github.com/Darkflame72/asyncpixel/actions?workflow=Tests">\n  <img src="https://github.com/Darkflame72/asyncpixel/workflows/Tests/badge.svg" alt="Test status" />\n  </a>\n   <img src="https://img.shields.io/badge/Contributor%20Covenant-v2.0%20adopted-ff69b4.svg" alt="Code of conduct" />\n</p>\n\n<h3 align="center">\n  <a href="https://asyncpixel.readthedocs.org">Docs</a>\n  <span> · </span>\n  <a href="https://github.com/Darkflame72/asyncpixel/discussions?discussions_q=category%3AIdeas">Feature request</a>\n  <span> · </span>\n  <a href="https://github.com/Darkflame72/asyncpixel/issues">Report a bug</a>\n  <span> · </span>\n  Support: <a href="https://github.com/Darkflame72/asyncpixel/discussions">Discussions</a>\n</h3>\n\n## ✨ Features\n\n- **Asyncronous.** Unlike other libraries Asyncpixel is fully asynchronous. This makes it perfect to use in your next discord bot or powerful website without having to worry about blocking.\n\n- **100% API coverage.** Asyncpixel is currently the only python library with full coverage of the Hypixel API meaning that no endpoints are left untouched and out of your reach.\n\n- **Pydantic models.** All models are checked and validated by [Pydantic](https://github.com/samuelcolvin/pydantic) meaning that the data is always in the correct format perfect for you to use.\n\n- **Available on pypi.** Asyncpixel is available on pypi meaning no building from source, just use `pip install asyncpixel` to use it in your project.\n\n## 🏁 Getting Started with Asyncpixel\n\nTo get started check out the documentation which [lives here](https://asyncpixel.readthedocs.org) and is generously hosted by readthedocs.\n\n### Installation\n\nUse the package manager [pip](https://pip.pypa.io/en/stable/) or your favourite tool to install asyncpixel.\n\n```bash\npip install asyncpixel\n```\n\n### Example\n\n```python\nimport asyncpixel\nimport asyncio\n\nasync def main():\n  hypixel = asyncpixel.Hypixel("hypixel_api_key")\n  player = await hypixel.player("405dcf08b80f4e23b97d943ad93d14fd")\n  print(player.stats.bedwars)\n  await hypixel.close()\n\n\nasyncio.run(main())\n```\n\n## ❗ Code of Conduct\n\nDarkflame72 is dedicated to providing a welcoming, diverse, and harassment-free experience for everyone. We expect everyone in the Darkflame72 community to abide by our [**Code of Conduct**](https://github.com/Darkflame72/asyncpixel/blob/main/CODE_OF_CONDUCT.rst). Please read it.\n\n## 🙌 Contributing to Asyncpixel\n\nFrom opening a bug report to creating a pull request: every contribution is appreciated and welcomed. If you are planning to implement a new feature or change the library please create an issue first. This way we can ensure your work is not in vain.\n\n### Not Sure Where to Start?\n\nA good place to start contributing, are the [Good first issues](https://github.com/Darkflame72/asyncpixel/labels/good%20first%20issue).\n\n## 📝 License\n\nAsyncpixel is open-source. The library is licensed [GPL v3](https://www.gnu.org/licenses/gpl-3.0.en.html).\n\n## 💬 Get in touch\n\nIf you have a question or would like to talk with other Asyncpixel users, please hop over to [Github discussions](https://github.com/Darkflame72/asyncpixel/discussions).\n\n## Contributors ✨\n\nThanks goes to these wonderful people ([emoji key](https://allcontributors.org/docs/en/emoji-key)):\n\n<!-- ALL-CONTRIBUTORS-LIST:START - Do not remove or modify this section -->\n<!-- prettier-ignore-start -->\n<!-- markdownlint-disable -->\n<table>\n  <tr>\n    <td align="center"><a href="https://quirky.codes/"><img src="https://avatars2.githubusercontent.com/u/35202521?v=4?s=100" width="100px;" alt=""/><br /><sub><b>AjayACST</b></sub></a><br /><a href="#maintenance-AjayACST" title="Maintenance">🚧</a></td>\n    <td align="center"><a href="https://github.com/aiexz"><img src="https://avatars3.githubusercontent.com/u/42418433?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Alex</b></sub></a><br /><a href="https://github.com/Darkflame72/asyncpixel/commits?author=aiexz" title="Code">💻</a> <a href="https://github.com/Darkflame72/asyncpixel/issues?q=author%3Aaiexz" title="Bug reports">🐛</a></td>\n    <td align="center"><a href="https://github.com/magicaltoast"><img src="https://avatars.githubusercontent.com/u/68669235?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Damian Grzanka</b></sub></a><br /><a href="https://github.com/Darkflame72/asyncpixel/commits?author=magicaltoast" title="Code">💻</a></td>\n    <td align="center"><a href="https://github.com/satyamedh"><img src="https://avatars.githubusercontent.com/u/47636284?v=4?s=100" width="100px;" alt=""/><br /><sub><b>satyamedh hulyalkar</b></sub></a><br /><a href="https://github.com/Darkflame72/asyncpixel/issues?q=author%3Asatyamedh" title="Bug reports">🐛</a></td>\n    <td align="center"><a href="https://dubs.rip"><img src="https://avatars.githubusercontent.com/u/59372145?v=4?s=100" width="100px;" alt=""/><br /><sub><b>dubs</b></sub></a><br /><a href="https://github.com/Darkflame72/asyncpixel/issues?q=author%3Aduhby" title="Bug reports">🐛</a></td>\n    <td align="center"><a href="https://github.com/Amund211"><img src="https://avatars.githubusercontent.com/u/14028449?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Amund Eggen Svandal</b></sub></a><br /><a href="https://github.com/Darkflame72/asyncpixel/commits?author=Amund211" title="Code">💻</a> <a href="https://github.com/Darkflame72/asyncpixel/issues?q=author%3AAmund211" title="Bug reports">🐛</a></td>\n  </tr>\n</table>\n\n<!-- markdownlint-restore -->\n<!-- prettier-ignore-end -->\n\n<!-- ALL-CONTRIBUTORS-LIST:END -->\n\nThis project follows the [all-contributors](https://github.com/all-contributors/all-contributors) specification. Contributions of any kind are welcome!\n',
    'author': 'Darkflame72',
    'author_email': 'leon@bowie-co.nz',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Darkflame72/asyncpixel',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""A Python HypixelAPI wrapper."""
#  Copyright (C) 2020 Leon Bowie
# This program is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License a
# long with this program. If not, see <https://www.gnu.org/licenses/>.
import datetime
import uuid
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import aiohttp
from pydantic import parse_obj_as

from .exceptions import ApiNoSuccessError
from .exceptions import InvalidApiKeyError
from .exceptions import RateLimitError
from .models import Auction
from .models import AuctionItem
from .models import Bazaar
from .models import BazaarItem
from .models import Boosters
from .models import Friend
from .models import Game
from .models import GameCounts
from .models import Guild
from .models import Key
from .models import Leaderboards
from .models import News
from .models import Player
from .models import Profile
from .models import Status
from .models import WatchDog
from .utils import calc_player_level

__all__ = ["Hypixel"]

BASE_URL = "https://api.hypixel.net/"

UUID = Union[str, uuid.UUID]


class Hypixel:
    """Client class for hypixel wrapper."""

    def __init__(
        self,
        api_key: Optional[UUID] = None,
    ) -> None:
        """Initialise client object.

        Args:
            api_key (Optional[UUID], optional): hypixel api key. Defaults to None.
        """
        if isinstance(api_key, str) and api_key is not None:
            api_key = uuid.UUID(api_key)
        self.api_key = api_key
        self._session = aiohttp.ClientSession()
        self.requests_remaining: int = -1
        self.total_requests: int = 0
        self._ratelimit_reset: datetime.datetime = datetime.datetime(1998, 1, 1)
        self.retry_after: datetime.datetime = datetime.datetime(1998, 1, 1)
        self._calc_player_level = calc_player_level

    async def close(self) -> None:
        """Used for safe client cleanup."""
        await self._session.close()

    async def _get(
        self,
        path: str,
        params: Optional[Dict[str, Any]] = None,
        key_required: bool = True,
    ) -> Dict[str, Any]:
        """Base function to get raw data from hypixel.

        Args:
            path (str):
                path that you wish to request from.
            params (Dict, optional):
                parameters to pass into request defaults to empty dictionary.
            key_required (bool): Whether an api key is needed

        Raises:
            RateLimitError: error if ratelimit has been reached.
            InvalidApiKeyError: error if api key is invalid.
            ApiNoSuccessError: error if api throughs an error.

        Returns:
            dict: returns a dictionary of the json response.
        """
        if (
            self.requests_remaining != -1
            and (
                self.requests_remaining == 0
                and self._ratelimit_reset > datetime.datetime.now()
            )
            or self.retry_after
            and (self.retry_after > datetime.datetime.now())
        ):
            raise RateLimitError(self.retry_after)

        if params is None:
            params = {}
        if key_required:
            if self.api_key is None:
                raise InvalidApiKeyError("No API key provided")
            params["key"] = str(self.api_key)

        response: aiohttp.ClientResponse = await self._session.get(
            f"{BASE_URL}{path}", params=params
        )

        if response.status == 429:
            self.requests_remaining = 0
            self.retry_after = datetime.datetime.now() + datetime.timedelta(
                seconds=int(response.headers["Retry-After"])
            )
            raise RateLimitError(
                datetime.datetime.now()
                + datetime.timedelta(seconds=int(response.headers["Retry-After"]))
            )

        elif response.status == 403:
            raise InvalidApiKeyError()

        elif response.status != 200:
            raise ApiNoSuccessError(path)

        elif key_required and "RateLimit-Limit" in response.headers:
            if self.total_requests == 0:
                self.total_requests = int(response.headers["RateLimit-Limit"])
            self.requests_remaining = int(response.headers["RateLimit-Remaining"])
            self._ratelimit_reset = datetime.datetime.now() + datetime.timedelta(
                seconds=int(response.headers["RateLimit-Reset"])
            )

        data: Dict[str, Any] = await response.json()

        return data

    async def watchdog_stats(self) -> WatchDog:
        """Get current watchdog stats.

        Returns:
            WatchDog: WatchDog stats object.
        """
        data = await self._get("watchdogstats")

        return WatchDog.parse_obj(data)

    async def key_data(self, key: Optional[str] = None) -> Key:
        """Get information about an api key.

        Args:
            key (str, optional): api key. Defaults token provided in class.

        Raises:
            InvalidApiKeyError: No api key available.

        Returns:
            Key: Key object.
        """
        if not key:
            key = str(self.api_key)

        if key == "None":
            raise InvalidApiKeyError()

        params = {"key": key}

        data = (await self._get("key", params=params, key_required=False))["record"]

        return Key.parse_obj(data)

    async def boosters(self) -> Boosters:
        """Get the current online boosters.

        Returns:
            Boosters: object containing boosters.
        """
        data = await self._get("boosters")
        data["decrementing"] = data["boosterState"]["decrementing"]
        return Boosters.parse_obj(data)

    async def player_count(self) -> int:
        """Get the current amount of players online.

        Returns:
            int: number of online players.
        """
        data = await self._get("playerCount")

        return int(data["playerCount"])

    async def news(self) -> List[News]:
        """Get current skyblock news.

        Returns:
            List[News]: List of news objects.
        """
        data = await self._get("skyblock/news")

        return parse_obj_as(List[News], data["items"])

    async def player_status(self, uuid: UUID) -> Optional[Status]:
        """Get current online status about a player.

        Args:
            uuid (UUID): uuid of player.

        Returns:
            Status: Status object of player.
        """
        data = await self._get("status", params={"uuid": str(uuid)})
        if data["session"] is None:
            return None
        return Status.parse_obj(data["session"])

    async def player_friends(self, uuid: UUID) -> Optional[List[Friend]]:
        """Get a list of a players friends.

        Args:
            uuid (UUID): the uuid of the player you wish to get friends from.

        Returns:
            List[Friend]: returns a list of friend elements.
        """
        params = {"uuid": str(uuid)}
        data = await self._get("friends", params=params)
        if data["records"] is None:
            return None

        return parse_obj_as(List[Friend], data["records"])

    async def bazaar(self) -> Bazaar:
        """Get info of the items in the bazaar.

        Returns:
            Bazaar: object for bazzar.
        """
        data = await self._get("skyblock/bazaar")

        bazaar_items = []

        for name in data["products"]:
            elements = data["products"][name]
            elements["name"] = name
            bazaar_items.append(BazaarItem.parse_obj(elements))

        return Bazaar(
            last_updated=data["lastUpdated"],
            bazaar_items=bazaar_items,
        )

    async def auctions(self, page: int = 0, retry: int = 3) -> Auction:
        """Get the auctions available.

        Args:
            page (int): Page of auction list you want. Defaults to 0.
            retry (int): Amount of retries to get the data from the api Defaults to 3.

        Returns:
            Auction: Auction object.
        """
        params = {"page": page}
        for _ in range(retry):  # pragma: no cover
            try:
                data = await self._get("skyblock/auctions", params=params)
                break
            except aiohttp.ServerTimeoutError:
                pass
        return Auction.parse_obj(data)

    async def recent_games(self, uuid: UUID) -> Optional[List[Game]]:
        """Get recent games of a player.

        Args:
            uuid (UUID): uuid of player.

        Returns:
            List[Game]: list of recent games.
        """
        params = {"uuid": str(uuid)}
        data = await self._get("recentGames", params=params)
        if data["games"] is None:
            return None

        return parse_obj_as(List[Game], data["games"])

    async def player(self, uuid: UUID) -> Optional[Player]:
        """Get information about a player from their uuid.

        Args:
            uuid (UUID): uuid of player.

        Returns:
            Player: player object.
        """
        params = {"uuid": str(uuid)}
        data = await self._get("player", params=params)
        if data["player"] is None:
            return None
        return Player.parse_obj(data["player"])

    async def guild_by_name(self, guild_name: str) -> Optional[Guild]:
        """Get guild by name.

        Args:
            guild_name (str): name of guild.

        Returns:
            Guild: guild object.
        """
        params = {"name": guild_name}
        data = await self._get("guild", params=params)
        if data["guild"] is None:
            return None
        return Guild.parse_obj(data["guild"])

    async def guild_by_id(self, guild_id: str) -> Optional[Guild]:
        """Get guild by id.

        Args:
            guild_id (str): id of guild.

        Returns:
            Guild: guild object.
        """
        params = {"id": guild_id}
        data = await self._get("guild", params=params)
        if data["guild"] is None:
            return None
        return Guild.parse_obj(data["guild"])

    async def guild_by_player(self, player_uuid: UUID) -> Optional[Guild]:
        """Get guild by player.

        Args:
            player_uuid (UUID): uuid of a player in the guild.

        Returns:
            Guild: guild object.
        """
        params = {"player": str(player_uuid)}
        data = await self._get("guild", params=params)
        if data["guild"] is None:
            return None
        return Guild.parse_obj(data["guild"])

    async def auction_from_uuid(self, uuid: UUID) -> Optional[List[AuctionItem]]:
        """Get auction from uuid.

        Args:
            uuid (UUID): minecraft uuid.

        Returns:
            List[AuctionItem]: list of auctions.
        """
        params = {"uuid": str(uuid)}
        data = await self._get("skyblock/auction", params=params)
        if data["auctions"] is None:
            return None
        return parse_obj_as(List[AuctionItem], data["auctions"])

    async def auction_from_player(self, player: str) -> Optional[List[AuctionItem]]:
        """Get auction data from player.

        Args:
            player (str): player.

        Returns:
            List[AuctionItem]: list of auction items.
        """
        params = {"player": player}
        data = await self._get("skyblock/auction", params=params)
        if data["auctions"] is None:
            return None
        return parse_obj_as(List[AuctionItem], data["auctions"])

    async def auction_from_profile(
        self, profile_id: str
    ) -> Optional[List[AuctionItem]]:
        """Get auction data from profile.

        Args:
            profile_id (str): profile id.

        Returns:
            List[AuctionItem]: list of auction items.
        """
        params = {"profile": profile_id}
        data = await self._get("skyblock/auction", params=params)
        if data["auctions"] is None:
            return None
        return parse_obj_as(List[AuctionItem], data["auctions"])

    async def game_count(self) -> GameCounts:
        """Gets number of players per game.

        Returns:
            GameCounts: game counts.
        """
        data = await self._get("gameCounts")
        return GameCounts.parse_obj(data)

    async def leaderboards(self) -> Dict[str, List[Leaderboards]]:
        """Get the current leaderboards.

        Returns:
            Dict[str, Leaderboards]: raw json response.
        """
        data = await self._get("leaderboards")
        leaderboard = {}
        leaderboards = data["leaderboards"]
        for key in leaderboards.keys():
            leaderboards_list = []
            for element in leaderboards[key]:
                location: Tuple[int, int, int] = (
                    int(element["location"].split(",")[0]),
                    int(element["location"].split(",")[1]),
                    int(element["location"].split(",")[2]),
                )
                leaderboards_list.append(
                    Leaderboards(
                        path=element["path"],
                        prefix=element["prefix"],
                        title=element["title"],
                        location=location,
                        count=element["count"],
                        leaders=element["leaders"],
                    )
                )
            leaderboard[key] = leaderboards_list

        return leaderboard

    async def resources_achievements(self) -> Dict[str, Any]:
        """Get the current resources. Does not require api key.

        Returns:
            Dict[str, Any]: raw json response.
        """
        data = await self._get("resources/achievements", key_required=False)
        return data

    async def resources_challenges(self) -> Dict[str, Any]:
        """Get the current resources. Does not require api key.

        Returns:
            Dict[str, Any]: raw json response.
        """
        data = await self._get("resources/challenges", key_required=False)
        return data

    async def resources_quests(self) -> Dict[str, Any]:
        """Get the current resources. Does not require api key.

        Returns:
            Dict[str, Any]: raw json response.
        """
        data = await self._get("resources/quests", key_required=False)
        return data

    async def resources_guilds_achievements(self) -> Dict[str, Any]:
        """Get the current resources. Does not require api key.

        Returns:
            Dict[str, Any]: raw json response.
        """
        data = await self._get("resources/guilds/achievements", key_required=False)
        return data

    async def resources_guilds_permissions(self) -> Dict[str, Any]:
        """Get the current resources. Does not require api key.

        Returns:
            Dict[str, Any]: raw json response.
        """
        data = await self._get("resources/guilds/permissions", key_required=False)
        return data

    async def resources_skyblock_collections(self) -> Dict[str, Any]:
        """Get the current resources. Does not require api key.

        Returns:
            Dict[str, Any]: raw json response.
        """
        data = await self._get("resources/skyblock/collections", key_required=False)
        return data

    async def resources_skyblock_skills(self) -> Dict[str, Any]:
        """Get the current resources. Does not require api key.

        Returns:
            Dict[str, Any]: raw json response.
        """
        data = await self._get("resources/skyblock/skills", key_required=False)
        return data

    async def profile(self, profile: str) -> Optional[Profile]:
        """Get profile info of a skyblock player.

        Args:
            profile (str): profile id of player can be gotten from
                            running profiles.

        Returns:
            Union[Profile, None]: Profile if it exists
        """
        params = {"profile": profile}
        data = await self._get("skyblock/profile", params=params)
        if data["profile"] is None:
            return None
        return Profile.parse_obj(data["profile"])

    async def profiles(self, uuid: UUID) -> Optional[Dict[str, Profile]]:
        """Get info on a profile.

        Args:
            uuid (UUID): uuid of player.

        Returns:
            Dict[str, Profile]: json response.
        """
        params = {"uuid": str(uuid)}
        data = await self._get("skyblock/profiles", params=params)
        if data["profiles"] is None:
            return None
        profiles = data["profiles"]
        profile_dict = {}
        for profile in profiles:
            _id = profile["profile_id"]
            profile_dict[_id] = Profile.parse_obj(profile)
        return profile_dict

    async def uuid_from_name(self, username: str) -> Optional[uuid.UUID]:
        """Helper method to get uuid from username.

        Args:
            username (str): username of player

        Returns:
            UUID4: uuid of player
        """
        async with self._session.get(
            "https://api.mojang.com/users/profiles/minecraft/" f"{username}"
        ) as response:
            if response.status != 200:
                return None
            else:
                return uuid.UUID((await response.json())["id"])



"""Constants."""
import json
import os
from functools import lru_cache
from typing import Any
from typing import Dict
from typing import List

from .models.game_type import GameType

script_dir = os.path.dirname(os.path.realpath(__file__))

# with open("asyncpixel/hypixelconstants/build/achievements.json") as file:
#     achievements: Dict[str, Any] = json.load(file)

# with open("asyncpixel/hypixelconstants/build/achievements_extended.json") as file:
#     achivements_extended: Dict[str, Any] = json.load(file)

# with open("asyncpixel/hypixelconstants/build/challenges.json") as file:
#     challenges: Dict[str, Any] = json.load(file)

# with open("asyncpixel/hypixelconstants/build/guild_achivements.json") as file:
#     guild_achivements: Dict[str, Any] = json.load(file)


# with open("asyncpixel/hypixelconstants/build/languages.json") as file:
#     languages: Dict[str, Any] = json.load(file)

# with open("asyncpixel/hypixelconstants/build/maps.json") as file:
#     maps: Dict[str, Any] = json.load(file)

# with open("asyncpixel/hypixelconstants/build/modes.json") as file:
#     modes: Dict[str, Any] = json.load(file)

# with open("asyncpixel/hypixelconstants/build/pet_xp.json") as file:
#     pet_xp: Dict[str, Any] = json.load(file)

# with open("asyncpixel/hypixelconstants/build/quests.json") as file:
#     quests: Dict[str, Any] = json.load(file)

# with open("asyncpixel/hypixelconstants/build/skyblock_bazaar.json") as file:
#     skyblock_bazaar: Dict[str, Any] = json.load(file)

# with open("asyncpixel/hypixelconstants/build/skyblock_collections.json") as file:
#     skyblock_collections: Dict[str, Any] = json.load(file)

# with open("asyncpixel/hypixelconstants/build/skyblock_items.json") as file:
#     skyblock_items: Dict[str, Any] = json.load(file)

# with open("asyncpixel/hypixelconstants/build/achievements.json") as file:
#     achievements: Dict[str, Any] = json.load(file)

# with open("asyncpixel/hypixelconstants/build/skyblock_skills.json") as file:
#     skyblock_skills: Dict[str, Any] = json.load(file)


@lru_cache()
def get_game_types() -> List[GameType]:
    """Get current game types.

    Returns:
        List[GameType]: game types.
    """
    abs_file_path = os.path.join(script_dir, "hypixelconstants/build/game_types.json")
    with open(abs_file_path) as file:
        game_types_data: Dict[str, Any] = json.load(file)
    return [GameType.parse_obj(data) for data in game_types_data]



"""TNT Games."""
from pydantic import BaseModel
from pydantic import Field


class TNTGames(BaseModel):
    """TNT games stats.

    Args:
        coins (int): Number of coins gathered in this Game Mode. Defaults to 0.
        deaths (int): Total deaths. Defaults to 0.
        wins (int): Total wins. Defaults to 0.
        record (int): Record. Defaults to 0.
        winstreak (int): Current winstreak. Defaults to 0.
    """

    coins: int = 0
    deaths: int = Field(0, alias="deaths_tntrun")
    wins: int = 0
    record: int = Field(0, alias="record_tntrun")
    winstreak: int = 0



"""Hunger Games."""
from pydantic import BaseModel


class HungerGames(BaseModel):
    """Hunger Games games stats.

    Args:
        coins (int): Number of coins gathered in this Game Mode. Defaults to 0.
        deaths (int): Total deaths. Defaults to 0.
        damge (int): Total damage dealt. Defaults to 0.
        wins (int): Total wins. Defaults to 0.
        games_played (int): Total games player. Defaults to 0.
    """

    coins: int = 0
    deaths: int = 0
    damge: int = 0
    wins: int = 0
    games_played: int = 0



"""Guild objects."""
import datetime
import uuid
from typing import Dict
from typing import List
from typing import Optional

from asyncpixel.models.utils import to_camel
from pydantic import BaseModel
from pydantic.fields import Field


class Pattern(BaseModel):
    """Pattern.

    Args:
        color (int): colour.
        pattern (str): pattern.
    """

    color: int
    pattern: str


class GuildMembers(BaseModel):
    """Members in a guild.

    Args:
        uuid (uuid.UUID): UUID of player.
        rank (str): Rank of player.
        joined (datetime.datetime): Time player joined guild.
        exp_history (Dict[str, int]): Exp history of player.
        quest_participation (int): How many quests the player has participated in.
        muted_till (Optional[datetime.datetime]): Time player unmuted. Defaults to None.
    """

    uuid: uuid.UUID
    rank: str
    joined: datetime.datetime
    exp_history: Optional[Dict[str, int]]
    quest_participation: Optional[int] = None
    muted_till: Optional[datetime.datetime] = None

    class Config:
        """Config."""

        alias_generator = to_camel


class Rank(BaseModel):
    """Rank.

    Args:
        name (str): Name of rank.
        default (bool): wether its the default.
        created (int): Created.
        priority (int): Priority of role.
        tag (str): Tag of role.
    """

    name: str
    default: bool
    created: int
    priority: int
    tag: Optional[str] = None


class Guild(BaseModel):
    """Guild object.

    Args:
        id (str): Guild ID.
        created (datetime.datetime): Timestamp this guild was created at.
        name (str): Name of guild.
        name_lower (str: Priority of role.
        description (str): Description of this guild that appears in the
            guild list and /g info.
        tag (str): Tag of guild.
        exp (int): Exp or guild.
        members (List[GuildMembers]): Array of guild members.
        achievements (Dict[str, int]): Guild achievements earned and
            the current progress.
        ranks (List[Rank]): Array of guild ranks.
        joinable (bool): Whether this guild can be joined using /g join.
        legacy_ranking (int): Ranking in the number of guild coins owned in
            the legacy guild system (0-indexed).
        publicly_listed (bool): Whether this guild is listed in the Guild Finder.
        preferred_games (List[str]): This guild's set preferred games.
        chat_mute (datetime.datetime): Timestamp guild chat will be unmuted at,
            or 0 if guild chat is not muted.
        guild_exp_by_game_type (Dict[str, str]): Amount of EXP earned for this guild
            by which game it was earned in.
        tag_color (Optional[str]): Color of this guild's tag, if set. Defaults to None.
    """

    id: str = Field(alias="_id")
    created: datetime.datetime
    name: str
    name_lower: str = Field(alias="name_lower")
    description: Optional[str]
    tag: Optional[str]
    exp: int
    members: List[GuildMembers]
    achievements: Dict[str, int]
    ranks: Optional[List[Rank]]
    joinable: bool = False
    legacy_ranking: Optional[int] = None
    publicly_listed: Optional[bool]
    preferred_games: Optional[List[str]]
    chat_mute: Optional[datetime.datetime] = None
    guild_exp_by_game_type: Optional[Dict[str, int]]
    tag_color: Optional[str]

    class Config:
        """Config."""

        alias_generator = to_camel



"""Skywars."""
from asyncpixel.models.utils import safe_divide
from pydantic import BaseModel
from pydantic import Field


class Skywars(BaseModel):
    """Skywars games stats.

    Args:
        coins (int): Number of coins gathered in this Game Mode. Defaults to 0.
        games_played (int): Total games played. Defaults to 0.
        tokens (int): Number of tokens. Defaults to 0.
        souls (int): Souls. Defaults to 0.
        winstreak (int): Current winstreak. Defaults to 0.
        kills (int): Total kills. Defaults to 0.
        deaths (int): Total deaths. Defaults to 0.
        wins (int): Total wins Defaults to 0.
        losses (int): Total losses. Defaults to 0.
    """

    coins: int = 0
    games_played: int = Field(0, alias="games_played_skywars")
    tokens: int = Field(0, alias="cosmetic_tokens")
    souls: int = 0
    winstreak: int = Field(0, alias="win_streak")
    kills: int = 0
    deaths: int = 0
    wins: int = 0
    losses: int = 0

    @property
    def kills_per_death(self) -> float:
        """Kills per deaths.

        Returns:
            float: ratio between kills and deaths.
        """
        return safe_divide(self.kills, self.deaths)

    @property
    def wins_per_lose(self) -> float:
        """Wins per losses..

        Returns:
            float: ratio between wins and losses.
        """
        return safe_divide(self.wins, self.losses)
